package at.ac.tuwien.infosys.jaxb.test;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.Date;
import java.util.List;

import javax.jws.WebService;
import javax.jws.soap.SOAPBinding;
import javax.jws.soap.SOAPBinding.ParameterStyle;
import javax.xml.bind.annotation.Annotation;
import javax.xml.bind.annotation.AppInfo;
import javax.xml.bind.annotation.Documentation;
import javax.xml.bind.annotation.Facets;
import javax.xml.bind.annotation.MaxOccurs;
import javax.xml.bind.annotation.MinOccurs;
import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlSchemaType;
import javax.xml.bind.annotation.XmlTransient;
import javax.xml.ws.Endpoint;

/**
 * @author Waldemar Hummer (hummer@infosys.tuwien.ac.at)
 */
@SuppressWarnings("all")
public class FacetsTestUtil {

	/* Thanks to Uwe Maurer for the following test class..! */
	@XmlAccessorType(XmlAccessType.FIELD)
	@SuppressWarnings("unused")
	@Documentation("A Buddy.")
	public static class Buddy {
	    @Documentation("Name of buddy.")
	    private String name;
	}
	
	@WebService
	public static interface TestWebService {
		void foo(TestRequest r);
	}
	@WebService
	public static class TestWebServiceImpl implements TestWebService {
		@SOAPBinding(parameterStyle=ParameterStyle.BARE)
		public void foo(TestRequest r) { }
	}

	public void testGenerateFacetsForAttributes() throws Exception {
		String url = "http://localhost:33452/service1";
		String wsdl = url + "?wsdl";
		TestWebServiceImpl s = new TestWebServiceImpl();
		Endpoint.publish(url, s);
		String c = readURL(wsdl);
		if(c.contains("schemaLocation=")) {
			String schema = c.substring(c.indexOf("schemaLocation="));
			schema = schema.substring(schema.indexOf("\"") + 1);
			schema = schema.substring(0, schema.indexOf("\""));
			c = readURL(schema);
		}
		System.out.println(c);
		System.out.println("sleeping 5 seconds...");
		Thread.sleep(5000);
	}
	private static String readURL(String url) throws Exception {
		return readStream(new URL(url).openStream());
	}
	private static String readStream(InputStream is) throws Exception {
		BufferedReader r = new BufferedReader(new InputStreamReader(is));
		String t = null;
		String c = "";
		while((t = r.readLine()) != null) {
			c += t + "\n";
		}
		return c;
	}
	
	public static void main(String[] args) throws Exception {
		FacetsTestUtil t = new FacetsTestUtil();
		t.testGenerateFacetsForAttributes();
		System.out.println("Press ENTER");
		System.in.read();
		System.exit(0);
	}
}
